"""
fabik is a library for Python command line application.
-----------------------
"""

__version__ = '25.7.1'
__author__ = 'zrong'
__name__ = 'fabik'
__email__ = 'zrongzrong@gmail.com'
__url__ = 'https://github.com/zrong/fabik'
__license__ = 'MIT'
__description__ = 'fabik is a library for Python command line application.'

import datetime
__now__ = datetime.datetime.now()

__all__ = [
    '__version__',
    '__author__',
    '__email__',
    '__url__',
    '__license__',
    '__description__',
    '__now__',
]

